import types
from _typeshed import Incomplete
from cf_extension_core.resource_base import ResourceBase as _ResourceBase
from cloudformation_cli_python_lib.interface import (
    BaseModel as BaseModel,
    BaseResourceHandlerRequest as _BaseResourceHandlerRequest,
)
from mypy_boto3_dynamodb.service_resource import DynamoDBServiceResource
from typing import Literal, Optional, Type, TYPE_CHECKING

logger: Incomplete

class ResourceUpdate(_ResourceBase):
    _updated_model: Incomplete
    _was_model_updated: bool
    def __init__(
        self,
        request: _BaseResourceHandlerRequest,
        db_resource: DynamoDBServiceResource,
        primary_identifier: str,
        type_name: str,
    ) -> None: ...
    def update_model(self, updated_model: _ResourceBase.T) -> None: ...
    def read_model(self, model_type: Type[_ResourceBase.T]) -> _ResourceBase.T: ...
    def __enter__(self) -> ResourceUpdate: ...
    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[types.TracebackType],
    ) -> Literal[False]: ...



from typing import Any, MutableMapping, Optional

class CustomResourceHelpers:
    ALL_HANDLER_TIMEOUT_THAT_SUPPORTS_IN_PROGRESS: int
    READ_LIST_HANDLER_TIMEOUT: int
    STANDARD_SEPARATOR: str
    HANDLER_ENTRY_TIME: Any

    @staticmethod
    def init_logging() -> None: ...
    @staticmethod
    def generate_id_resource(
        stack_id: Optional[str],
        logical_resource_id: Optional[str],
        resource_identifier: str,
    ) -> str: ...
    @staticmethod
    def generate_id_read_only_resource(
        stack_id: Optional[str],
        logical_resource_id: Optional[str],
    ) -> str: ...
    @staticmethod
    def get_naked_resource_identifier_from_string(primary_identifier: str) -> str: ...
    @staticmethod
    def generate_unique_resource_name(
        stack_id: str,
        logical_resource_name: str,
        client_request_token: str,
        max_length: int = 255,
    ) -> str: ...
    @staticmethod
    def _callback_add_handler_entry_time(callback_context: MutableMapping[str, Any]) -> None: ...
    @staticmethod
    def _callback_add_resource_end_time(
        callback_context: MutableMapping[str, Any],
        total_allowed_time_in_minutes: int,
    ) -> None: ...
    @staticmethod
    def should_return_in_progress_due_to_handler_timeout(callback_context: MutableMapping[str, Any]) -> bool: ...
    @staticmethod
    def _return_failure_due_to_timeout(callback_context: MutableMapping[str, Any]) -> None: ...



"""
Generates the dynamo table and makes sure all the settings are correct for resource creation/deletion
"""
import logging
import time
import botocore.exceptions
from typing import TYPE_CHECKING
from cf_extension_core.constants import DynamoDBValues

if TYPE_CHECKING:
    from mypy_boto3_dynamodb.service_resource import DynamoDBServiceResource
else:
    DynamoDBServiceResource = object

# Module Logger
logger = logging.getLogger(__name__)


class DynamoTableCreator:
    def __init__(self, db_resource: DynamoDBServiceResource):
        self._client: DynamoDBServiceResource = db_resource

    def delete_table(self) -> None:
        if self.table_exists():
            logger.info("Deleting table: " + DynamoDBValues.TABLE_NAME)
            the_table = self._client.Table(DynamoDBValues.TABLE_NAME)
            the_table.delete()
            self._wait_for_table_to_be_deleted()

    def table_exists(self) -> bool:
        try:
            self._client.meta.client.describe_table(TableName=DynamoDBValues.TABLE_NAME)
            logger.debug("Table Exists")
            return True
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                logger.debug("Table Does not Exist")
                return False
            else:
                raise

    def create_standard_table(self) -> None:

        logger.debug("In create_standard_table")

        if self.table_exists():
            logger.debug("Exit create_standard_table")
            return

        self._create_table(
            name=DynamoDBValues.TABLE_NAME,
            partition_key=DynamoDBValues.PARTITION_KEY,
        )
        logger.debug("Exit create_standard_table")

    def _create_table(
        self,
        name: str,
        partition_key: str,
    ) -> None:
        """
        Guarantees to create the table in dynamo db and waits for it to be active.
        If table already exists, this code exits cleanly
        """
        logger.debug("Creating table: " + name)

        try:
            self._client.create_table(
                AttributeDefinitions=[{"AttributeName": partition_key, "AttributeType": "S"}],
                KeySchema=[{"AttributeName": partition_key, "KeyType": "HASH"}],  # Required for a partition key
                TableName=name,
                BillingMode="PAY_PER_REQUEST",  # ON Demand
                SSESpecification={"Enabled": False},  # Opposite of what you think it should be.
            )

            # Wait for table to be active
            self._wait_for_table_to_be_active()

        except (
            self._client.meta.client.exceptions.ResourceInUseException,
            self._client.meta.client.exceptions.TableAlreadyExistsException,
        ) as ex:

            logger.debug("Table already exists")
            self._wait_for_table_to_be_active()

            # Table exists validate some of the properties?
            table = self._client.Table(name)

            # Key Schema
            keyschema = table.key_schema
            mypartkey = keyschema[0]
            if not (mypartkey["AttributeName"] == partition_key and mypartkey["KeyType"] == "HASH"):
                raise Exception("DynamoDB - Partition Key schema does not match: Actual: " + str(mypartkey)) from ex

            # If key schema is good, return
            # We are assuming here if the schema checks out, then this table should work for us.
            logger.debug("Table has correct settings, returning")
            return

    def _wait_for_table_to_be_active(self) -> None:
        logger.debug("Waiting till the table becomes ACTIVE")
        while True:
            status = self._client.Table(DynamoDBValues.TABLE_NAME).table_status
            if status == "ACTIVE":
                logger.debug("Table is Active")
                break
            else:
                time.sleep(2)

    def _wait_for_table_to_be_deleted(self) -> None:
        logger.debug("Waiting till the table deleted")
        while True:
            status = self.table_exists()
            if status is False:
                logger.debug("Table deleted")
                break
            else:
                time.sleep(2)


if __name__ == "__main__":
    import boto3

    boto3.setup_default_session(profile_name="CT", region_name="eu-west-2")
    dynamodb = boto3.resource("dynamodb")

    test = DynamoTableCreator(db_resource=dynamodb)
    print("Deleting table")
    test.delete_table()
    print("Creating table")
    test.create_standard_table()
    print("Table Created")



import logging
from cloudformation_cli_python_lib.interface import BaseResourceHandlerRequest as _BaseResourceHandlerRequest
from cloudformation_cli_python_lib.boto3_proxy import SessionProxy as _SessionProxy
from typing import TYPE_CHECKING, Optional, MutableMapping, Any

if TYPE_CHECKING:
    from mypy_boto3_dynamodb.service_resource import (
        DynamoDBServiceResource as _DynamoDBServiceResource,
    )
else:
    _DynamoDBServiceResource = object

# Locals
import cf_extension_core.resource_update as _resource_update
import cf_extension_core.resource_create as _resource_create
import cf_extension_core.resource_read as _resource_read
import cf_extension_core.resource_delete as _resource_delete
import cf_extension_core.resource_list as _resource_list
from cf_extension_core.dynamo_table_creator import DynamoTableCreator  # noqa: F401
from cf_extension_core.custom_resource_helpers import CustomResourceHelpers  # noqa: F401
from cf_extension_core.constants import DynamoDBValues  # noqa: F401

LOG = logging.getLogger(__name__)


def generate_dynamodb_resource(session_proxy: Optional[_SessionProxy]) -> _DynamoDBServiceResource:
    return session_proxy.resource(service_name="dynamodb")


def create_resource(
    request: _BaseResourceHandlerRequest,
    type_name: str,
    db_resource: _DynamoDBServiceResource,
) -> _resource_create.ResourceCreate:

    return _resource_create.ResourceCreate(
        db_resource=db_resource,
        type_name=type_name,
        request=request,
    )


def update_resource(
    primary_identifier: str,
    type_name: str,
    request: _BaseResourceHandlerRequest,
    db_resource: _DynamoDBServiceResource,
) -> _resource_update.ResourceUpdate:
    return _resource_update.ResourceUpdate(
        db_resource=db_resource,
        type_name=type_name,
        primary_identifier=primary_identifier,
        request=request,
    )


def delete_resource(
    primary_identifier: str,
    type_name: str,
    request: _BaseResourceHandlerRequest,
    db_resource: _DynamoDBServiceResource,
) -> _resource_delete.ResourceDelete:
    return _resource_delete.ResourceDelete(
        db_resource=db_resource,
        type_name=type_name,
        primary_identifier=primary_identifier,
        request=request,
    )


def read_resource(
    primary_identifier: str,
    type_name: str,
    request: _BaseResourceHandlerRequest,
    db_resource: _DynamoDBServiceResource,
) -> _resource_read.ResourceRead:
    return _resource_read.ResourceRead(
        db_resource=db_resource,
        type_name=type_name,
        primary_identifier=primary_identifier,
        request=request,
    )


def list_resource(
    type_name: str,
    request: _BaseResourceHandlerRequest,
    db_resource: _DynamoDBServiceResource,
) -> _resource_list.ResourceList:

    return _resource_list.ResourceList(db_resource=db_resource, type_name=type_name, request=request)


def initialize_handler(
    callback_context: MutableMapping[str, Any],
    total_allowed_time_in_minutes: int,
) -> None:
    LOG.debug("Start initialize_handler")

    # TODO: Consider overriding the Table name based on Type Name here
    CustomResourceHelpers._callback_add_resource_end_time(
        callback_context=callback_context,
        total_allowed_time_in_minutes=total_allowed_time_in_minutes,
    )
    CustomResourceHelpers._callback_add_handler_entry_time(callback_context=callback_context)
    CustomResourceHelpers._return_failure_due_to_timeout(callback_context=callback_context)

    LOG.debug("End initialize_handler")


def package_logging_config(logging_level: int) -> None:
    """
    Helps setup default logging config for custom resources
    :return:
    """

    logging.getLogger("cf_extension_core").setLevel(logging_level)
    LOG.info("cf_extension_core logging enabled")
